import appdirs
import diskcache
from time import sleep, time
import ib_insync
import yfinance as yf
import pandas_datareader.data as web

import datetime
from pandas import DataFrame, date_range, Series
import numpy as np

class Downloader:
    def __init__(self, start, end, tickers) -> None:
        self.start = start
        self.end = end
        self.tickers = tickers
        self.last_dates_downloaded = None

    def _calculate_last_date_downloaded(self, closes: DataFrame) -> None:
        self.last_dates_downloaded = closes.apply(Series.last_valid_index)

class InteractiveBrokersDownloader(Downloader):
    def __init__(self, start, end, tickers, currencies, exchanges) -> None:
        super().__init__(start, end, tickers)
        self.currencies = currencies
        self.exchanges = exchanges
        self.ib = None
    
    def connect(self) -> None:
        """Instantiate connection Trader Workstation
        """
        self.ib = ib_insync.IB()
        self.ib.connect('127.0.0.1', 7497, clientId=1, readonly=True)
        user_data = appdirs.user_cache_dir('assetuniverse', 'interactivebrokersdownloader')
        self.cache = diskcache.Cache(user_data)
        self.cache_validity_seconds = 20*60

    def shutdown(self) -> None:
        """Close connections
        """
        if self.ib:
            self.ib.disconnect()
    
    def download(self):
        closes = DataFrame()
        if self.tickers:
            self.connect()
            for symbol, currency, exchange in zip(self.tickers, self.currencies, self.exchanges):
                # Load from cache if already exists
                key = f'{symbol}{currency}{exchange}'
                cached = self.cache.get(key, None)
                download = True
                data = None
                if cached:
                    last_downloaded = cached.get('last_downloaded', 0)
                    if last_downloaded > time() - self.cache_validity_seconds:
                        delay = time() - last_downloaded
                        print(f'IB TWS: Loaded {symbol} data from cache.\tWas last downloaded {round(delay/60, 1)} minutes ago...')
                        data = cached['data']
                        download = False
                
                # Download if cache did not have recent data
                if download:
                    print(f'IB TWS: Downloading {symbol} in {currency} currency from {exchange} exchange...')
                    data = self._download_cont_future(symbol, currency, exchange)
                    self.cache[key] = {
                        'data': data,
                        'last_downloaded': time()
                    }

                # Join with other closes
                if closes.empty:
                    closes = data
                else:
                    closes = closes.join(data)
            super()._calculate_last_date_downloaded(closes)
        return closes
    
    def _download_cont_future(self, symbol:str, currency:str, exchange:str) -> DataFrame:
        """Download historical data for a continuous futures contract

        Parameters
        ----------
        symbol : str
            Symbol of the continuous futures contract
        currency : str
            Base currency of the contract
        exchange : str
            Exchange to target

        Returns
        -------
        DataFrame
            Historical closing prices. Example:
                    date        open        high         low       close  volume  average  barCount
            0    2019-12-27  116.210938  116.437500  115.886719  116.433594    -1.0     -1.0        -1
            1    2019-12-30  116.074219  116.218750  116.074219  116.175781    -1.0     -1.0        -1
            2    2019-12-31  116.179688  116.207031  116.144531  116.207031    -1.0     -1.0        -1
            3    2020-01-02  116.378906  116.531250  116.378906  116.421875    -1.0     -1.0        -1
        
        Raises
        ------
        ValueError
            _description_
        """
        contract = ib_insync.ContFuture(
            symbol=symbol,
            exchange=exchange,
            currency=currency
            )
        days = (self.end - self.start).days + 1
        duration = f'{days} D'
        if days > 365:
            duration = f'{int(np.ceil(days/365))} Y'
        for i in range(5):
            bars = self.ib.reqHistoricalData(
                contract, 
                endDateTime=None, 
                durationStr=duration,
                barSizeSetting='1 day', 
                whatToShow='ADJUSTED_LAST', 
                useRTH=True
                )
            if bars:
                break
            if i < 4:
                sleep(2)
                print(f'IB TWS: Downloading {symbol} in {currency} currency from {exchange} exchange... try #{i+2}')
        bars = ib_insync.util.df(bars)
        bars = DataFrame(data=bars['close'].values, index=bars['date'], columns=[symbol,])
        return bars
        


class YahooFinanceDownloader(Downloader):
    """Download from Yahoo Finance
    """
    def download(self):
        closes = DataFrame()
        if len(self.tickers) == 0:
            raise ValueError('There are no tickers to download.')
        if self.end == datetime.date.today():
            data = yf.download(self.tickers, interval="1d", auto_adjust=True, prepost=False, threads=True,
                                start=self.start)
        else:
            data = yf.download(self.tickers, interval="1d", auto_adjust=True, prepost=False, threads=True,
                                start=self.start, end=self.end)
        if len(self.tickers) > 1:
            closes = DataFrame(data["Close"][self.tickers])
        else:
            closes = DataFrame(data["Close"])
            closes = closes.rename(columns={"Close": self.tickers[0]})
        super()._calculate_last_date_downloaded(closes)
        return closes


class FredDownloader(Downloader):
    """Download from FRED
    """
    def __init__(self, start, end, tickers) -> None:
        super().__init__(start, end, tickers)
        self.freddic = {'5-Year Breakeven Inflation Rate': 'T5YIE',
                        '10-Year Breakeven Inflation Rate': 'T10YIE',
                        'Fed Funds Rate': 'DFF',
                        '3-Month Treasury Constant Maturity Rate': 'GS3M',
                        '1-Year Treasury Constant Maturity Rate': 'DGS1'}        
        
    def download(self):
        closes = DataFrame()
        for ticker in self.tickers:
            closes = web.DataReader(
                self.freddic.get(ticker, ticker), 
                'fred', 
                self.start, 
                self.end+datetime.timedelta(1)
                )
            closes = closes.rename(columns={closes.columns[0]: ticker})

            # Reindex tickers that aren't updated often
            if ticker in ['Fed Funds Rate',]:
                idx = date_range(start=self.start, end=self.end, freq='D')
                closes = closes.reindex(idx)
                closes.fillna(method="ffill", inplace=True)
        super()._calculate_last_date_downloaded(closes)
        return closes


class OfflineDownloader(Downloader):
    """Generates random prices and returns for situations without internet access
    """
    def download(self):
        diff = self.end - self.start
        offlineSym = list()
        r = np.exp(np.random.randn(diff.days + 1, len(offlineSym))/100) + 0.00001
        closes = DataFrame(data=np.cumprod(r, axis=0), columns=offlineSym)
        closes["Date"] = date_range(self.start, self.end, freq='D')
        closes = closes.set_index('Date')
        super()._calculate_last_date_downloaded(closes)
        return closes



import datetime
from typing import Dict, List
import pandas as pd

VALID_DATA_SOURCES = ['Yahoo Finance', 'FRED', 'Offline', 'Interactive Brokers']
VALID_SEC_TYPES = ['Stock', 'Future']

class Asset:
    def __init__(self, 
        start: datetime.date = datetime.date.today(),
        end: datetime.date = datetime.date.today()-datetime.timedelta(days=180),
        ticker: str = 'AAPL',
        alternate_tickers: List[str] = [],
        exchange: str = 'SMART',
        currency: str = 'USD',
        sectype: str = None,
        readable_name: str = None,
        downloader_definition: Dict = {},
        data_source: str = 'Yahoo Finance'
        ) -> None:

        self.start = start
        self.end = end
        self.ticker = ticker
        self.alternate_tickers = alternate_tickers
        self.exchange = exchange
        self.currency = currency
        if not sectype:
            sectype = 'Stock'
        if sectype not in VALID_SEC_TYPES:
            raise ValueError(f'Secrurity type {sectype} is not valid. Must be one of: {VALID_SEC_TYPES}')
        self.sectype = sectype
        self.readable_name = readable_name or ticker
        self.display_name = f'{ticker} - {self.readable_name}'
        self.downloader_definition = downloader_definition
        if data_source not in VALID_DATA_SOURCES:
            raise ValueError(f'Data source {data_source} is not valid. Must be one of: {VALID_DATA_SOURCES}')
        self.data_source = data_source
        self.prices = pd.DataFrame()
        self.prices_normalized = pd.DataFrame()
        self.returns = pd.DataFrame()

    def assign_prices(self, prices: pd.Series) -> None:
        """Assign the unnormalized daily closing prices. This function will automatically
        cal`cu`late the returns and normalized prices.

        Parameters
        ----------
        prices : pd.Series
            Unnormalized daily closing prices.
        """
        self.prices = prices
        self.prices_normalized = self.prices/self.prices[0] # Normalize price to start from $1
        self.returns = self.prices.pct_change()
        self.returns = self.returns[1:]  # delete first row - pct_change() returns first row as NaN

    def __str__(self):
        result = f'{self.ticker}, '
        if len(self.alternate_tickers):
            alternate_names = ', '.join([t for t in self.alternate_tickers])
            result = result + f'\t\talternate names: {alternate_names},\t'
        result = result + f'\t\tdisplays as {self.display_name}'
        return result
    
    def __repr__(self):
        return f'Asset({self.start}, {self.end}, \'{self.ticker}\')'


#!/usr/bin/env python

import copy
import datetime
import numpy as np
from pandas import DataFrame, to_datetime, date_range
import pandas as pd
import plotly.express as px
from typing import List

# Use these when importing assetuniverse as a library
from .utils.asset import Asset
from .utils.downloaders import InteractiveBrokersDownloader, YahooFinanceDownloader, FredDownloader, OfflineDownloader

# Use these when running core.py as main
# from utils.asset import Asset
# from utils.downloaders import YahooFinanceDownloader, FredDownloader, OfflineDownloader

# from assetuniverse import Asset
# from assetuniverse.downloaders import YahooFinanceDownloader, FredDownloader, OfflineDownloader

class AssetUniverse:
    def __init__(self, start, end, assets: List[Asset], cashasset:Asset, offline=False, borrow_spread=1.5):
        self.start = start
        self.end = end
        self.assets = {asset.ticker: asset for asset in assets}
        self.offline = offline
        self.cashasset = cashasset#Asset(start=start, end=end, ticker='VFISX', display_name='Cash')
        self.borrowrate = Asset(start=start, end=end, ticker='Fed Funds Rate', readable_name='Borrow Rate', data_source='FRED')
        self.borrow_spread = borrow_spread      # Percentage points above Fed Funds Rate

    def download(self) -> None:
        """Download all price and return data for all assets
        """
        print('Downloading asset universe data... ', flush=True)
        
        # Separate tickers into separate downloader lists
        InteractiveBrokersTickers = [a.ticker for a in self.assets.values() if a.data_source == 'Interactive Brokers']
        InteractiveBrokersCurrencies = [a.currency for a in self.assets.values() if a.data_source == 'Interactive Brokers']
        InteractiveBrokersExchanges = [a.exchange for a in self.assets.values() if a.data_source == 'Interactive Brokers']
        YahooFinanceTickers = [a.ticker for a in self.assets.values() if a.data_source == 'Yahoo Finance']
        if self.cashasset.ticker:
            YahooFinanceTickers = YahooFinanceTickers + [self.cashasset.ticker]
        FredTickers = [a.ticker for a in self.assets.values() if a.data_source == 'FRED']
        FredTickers = FredTickers + [self.borrowrate.ticker]
        OfflineTickers = [a.ticker for a in self.assets.values() if a.data_source == 'Offline']

        # Download
        ibd = InteractiveBrokersDownloader(self.start, self.end, InteractiveBrokersTickers, InteractiveBrokersCurrencies, InteractiveBrokersExchanges)
        yfd = YahooFinanceDownloader(self.start, self.end, YahooFinanceTickers)
        fd = FredDownloader(self.start, self.end, FredTickers)
        od = OfflineDownloader(self.start, self.end, OfflineTickers)
        prices_list = [
            ibd.download(),
            yfd.download(),
            fd.download(),
            od.download()
        ]
        ibd.shutdown()

        # Join all closes together on same date axis
        joined_prices = self._join_prices(prices_list)
        self.last_dates_downloaded = pd.concat((ibd.last_dates_downloaded, yfd.last_dates_downloaded, fd.last_dates_downloaded, od.last_dates_downloaded))

        # Rename cash and borrow rate
        cashname = 'Cash'
        borrowname = 'Borrow Rate'
        joined_prices = joined_prices.rename(columns={self.borrowrate.ticker: borrowname})
        self.borrowrate.ticker = borrowname
        if self.cashasset.ticker:
            joined_prices = joined_prices.rename(columns={self.cashasset.ticker: cashname})
            self.cashasset.ticker = cashname

        # Add spread to borrow rate and convert to "borrow price"
        joined_prices.loc[:, self.borrowrate.ticker] = joined_prices.loc[:, self.borrowrate.ticker] + self.borrow_spread
        daily_borrow_rate = (1 + joined_prices.loc[:, self.borrowrate.ticker]/100)**(1/252)
        joined_prices.loc[:, self.borrowrate.ticker] = daily_borrow_rate.cumprod()

        # Assign prices to each asset individually
        for asset in self.assets.values():
            asset.assign_prices(joined_prices[asset.ticker])
        if self.cashasset.ticker:
            self.cashasset.assign_prices(joined_prices[self.cashasset.ticker])
        else:
            ones = joined_prices[self.borrowrate.ticker]*0 + 1.0
            ones = ones.rename(cashname)
            self.cashasset.assign_prices(ones)
            self.cashasset.ticker = cashname
        self.borrowrate.assign_prices(joined_prices[self.borrowrate.ticker])
        print('Done.', flush=True)

    def _join_prices(self, prices_list: List[pd.DataFrame]) -> pd.DataFrame:
        """Join the prices of the raw downloaded data to have the same dates.

        Parameters
        ----------
        prices : List[pd.DataFrame]
            List of prices

        Returns
        -------
        pd.DataFrame
            Joined prices on same date axis
        """
        # Join all closes together
        joined_prices = pd.DataFrame()
        for i, prices in enumerate(prices_list):
            if not prices.empty and joined_prices.empty:
                # Find first non empty prices
                joined_prices = prices
                continue
            if not joined_prices.empty:
                # Join prices with previous prices
                joined_prices = joined_prices.join(prices, how='inner')
        # closes = closes.join(annualBorrowRate, how='left')    # Do I need this for the borrow rate? Different than how='inner'

        # Forward fill all the NaNs and zeros
        joined_prices[joined_prices == 0] = np.nan
        joined_prices.fillna(method="ffill", inplace=True)
        joined_prices.dropna(axis=0, how="any", inplace=True)

        # # Forward-fill cash closes - Do I need this part?
        # idx = date_range(self.start, self.end)
        # closesCash = closesCash.reindex(idx, method='ffill')

        return joined_prices

    def tickers(self, include_cash=True, include_borrow_rate=True) -> List[str]:
        """Get a list of all ticker symbols

        Parameters
        ----------
        include_cash : bool, optional
            Include the cash asset in the result, by default True
        include_borrow_rate : bool, optional
            Include the borrow rate in the result, by default True

        Returns
        -------
        List[str]
            List of ticker symbols in the asset universe
        """
        tickers = list(self.assets.keys())
        if include_cash:
            tickers = tickers + [self.cashasset.ticker]
        if include_borrow_rate:
            tickers = tickers + [self.borrowrate.ticker]
        return tickers

    def returns(self, tickers: List[str]=[], start: datetime.date = None, end: datetime.date = None) -> DataFrame:
        """Get the daily returns between start and end dates.

        Parameters
        ----------
        tickers : List[str]
            Tickers to include, by default all tickers in the AssetUniverse
        start : datetime.date, optional
            Start date, by default the start date of the AssetUniverse
        end : datetime.date, optional
            End date, by default the start date of the AssetUniverse

        Returns
        -------
        DataFrame
            Selected daily returns
        """
        if len(tickers) == 0:
            tickers = list(self.assets.keys()) + [self.cashasset.ticker, self.borrowrate.ticker]
        selected_returns = []
        for ticker in tickers:
            if ticker in self.assets.keys():
                returns = self.assets[ticker].returns
            elif ticker == self.cashasset.ticker:
                returns = self.cashasset.returns
            elif ticker == self.borrowrate.ticker:
                returns = self.borrowrate.returns
            else:
                raise ValueError(f'Ticker {ticker} not in the asset universe: {self.tickers()}')
            selected_returns.append(returns)
        all_returns = pd.concat(selected_returns, axis=1, join='inner')
        return all_returns.loc[start:end,:]

    def prices(self, tickers: List[str]=[], start: datetime.date = None, end: datetime.date = None, normalize=True) -> DataFrame:
        """Get the daily prices between start and end dates.

        Parameters
        ----------
        tickers : List[str]
            Tickers to include, by default all tickers in the AssetUniverse
        start : datetime.date, optional
            Start date, by default the start date of the AssetUniverse
        end : datetime.date, optional
            End date, by default the end date of the AssetUniverse
        normalize : bool, optional
            Normalize start prices to $1, by default True

        Returns
        -------
        DataFrame
            Selected daily prices
        """
        if len(tickers) == 0:
            tickers = list(self.assets.keys()) + [self.cashasset.ticker, self.borrowrate.ticker]
        selected_prices = []
        if normalize == True:
            for ticker in tickers:
                if ticker in self.assets.keys():
                    prices_normalized = self.assets[ticker].prices_normalized
                elif ticker == self.cashasset.ticker:
                    prices_normalized = self.cashasset.prices_normalized
                elif ticker == self.borrowrate.ticker:
                    prices_normalized = self.borrowrate.prices_normalized
                else:
                    raise ValueError(f'Ticker {ticker} not in the asset universe: {self.tickers()}')
                selected_prices.append(prices_normalized)
        else:
            for ticker in tickers:
                if ticker in self.assets.keys():
                    prices = self.assets[ticker].prices
                elif ticker == self.cashasset.ticker:
                    prices = self.cashasset.prices
                elif ticker == self.borrowrate.ticker:
                    prices = self.borrowrate.prices
                else:
                    raise ValueError(f'Ticker {ticker} not in the asset universe: {self.tickers()}')
                selected_prices.append(prices)
        all_prices = pd.concat(selected_prices, axis=1, join='inner')
        return all_prices.loc[start:end,:]


    def delete(self, tickers: List[str] = []):
        """Delete the tickers from the asset universe

        Parameters
        ----------
        tickers : List[str], optional
            Tickers to delete, by default []
        """
        if not isinstance(tickers, list):
            raise TypeError('Provide a list of tickers to delete.')
        for ticker in tickers:
            if ticker == self.cashasset.ticker:
                raise ValueError(f'Cannot delete cash asset: {ticker}')
            if ticker == self.borrowrate.ticker:
                raise ValueError(f'Cannot delete borrow rate: {ticker}')
            try:
                self.assets.pop(ticker)
            except KeyError as exc:
                raise KeyError(f'Could not delete requested ticker {ticker} because it doesn\'t exist in the asset universe')


    def plot_prices(self, tickers: List[str] = [], start: datetime.date = None, end: datetime.date = None, normalize=True) -> None:
        """Plot asset prices over time

        Parameters
        ----------
        tickers : List[str]
            Tickers to include, by default all tickers in the AssetUniverse except the borrow rate
        start : datetime.date, optional
            Start date, by default the start date of the AssetUniverse
        end : datetime.date, optional
            End date, by default the end date of the AssetUniverse
        normalize : bool, optional
            Normalize start prices to $1, by default True
        """
        if len(tickers) == 0:
            tickers = self.tickers(include_borrow_rate=False)
        prices = self.prices(tickers, start, end, normalize)
        renames = {}
        for ticker in prices.columns:
            asset = self.assets.get(ticker, None)
            if ticker == self.cashasset.ticker:
                asset = self.cashasset
            if ticker == self.borrowrate.ticker:
                asset = self.borrowrate
            if asset:
                if asset.display_name:
                    renames[ticker] = asset.display_name
        if len(renames):
            prices = prices.rename(columns=renames)
        prices['Date'] = prices.index
        fig = px.line(prices, x="Date", y=prices.columns)
        # fig = px.line(prices, x="Date", y=prices.columns,
        #       hover_data={"Date": "|%B %d, %Y"})
        fig.update_yaxes(
            type='log'
        )
        # update y axis name
        fig.update_yaxes(title_text='Price (Normalized)' if normalize else 'Price')
        fig.show()


    def __add__(self, other):
        # Combine two asset universes into one asset universe
        combinedAU = copy.deepcopy(self)
        if other is not None:
            # Check for compatibility
            if self.start != other.start or self.end != other.end:
                raise RuntimeError('Start or end dates do not match!')
            if self.offline != other.offline:
                raise RuntimeError('Offline statuses do not match!')

            # Merge symbols, prices, and returns
            combinedAU.sym = combinedAU.sym + other.sym
            combinedAU.p = combinedAU.p.join(other.p, how='inner')  # merge to have same dates
            combinedAU.r = combinedAU.r.join(other.r, how='inner')
            combinedAU.originalprices = combinedAU.originalprices.join(other.originalprices, how='inner')
        return combinedAU


    def correlation_matrix(self, tickers: List[str] = [], start: datetime.date = None, end: datetime.date = None) -> np.ndarray:
        """Calculate the correlation matrix

        Parameters
        ----------
        tickers : List[str]
            Tickers to include, by default all non-cash assets
        start : datetime.date, optional
            Start date, by default the start date of the AssetUniverse
        end : datetime.date, optional
            End date, by default the end date of the AssetUniverse

        Returns
        -------
        np.ndarray
            Correlation matrix
        """
        if len(tickers) == 0:
            tickers = self.tickers(include_cash=False, include_borrow_rate=False)
        returns = self.returns(tickers, start, end)
        return returns.corr(method='pearson')


    def covariance_matrix(self, tickers: List[str] = [], start: datetime.date = None, end: datetime.date = None) -> np.ndarray:
        """Calculate the correlation matrix

        Parameters
        ----------
        tickers : List[str]
            Tickers to include, by default all non-cash assets
        start : datetime.date, optional
            Start date, by default the start date of the AssetUniverse
        end : datetime.date, optional
            End date, by default the end date of the AssetUniverse

        Returns
        -------
        np.ndarray
            Covariance matrix
        """
        if len(tickers) == 0:
            tickers = self.tickers(include_cash=False, include_borrow_rate=False)
        returns = self.returns(tickers, start, end)
        return returns.cov()
